# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation, LevelGenerator
from minihack.envs import register


class MiniHackRiver(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 350)
        n_monster = kwargs.pop("n_monster", 0)
        n_boulder = kwargs.pop("n_boulder", 5)
        narrow = kwargs.pop("narrow", False)
        lava = kwargs.pop("lava", False)

        if narrow:
            map = """
..................WW.....
..................WW.....
..................WW.....
..................WW.....
..................WW.....
..................WW.....
..................WW.....
"""
        elif not lava:
            map = """
..................WWW....
..................WWW....
..................WWW....
..................WWW....
..................WWW....
..................WWW....
..................WWW....
"""
        else:
            map = """
..................LLL....
..................LLL....
..................WWW....
..................LLL....
..................WWW....
..................LLL....
..................LLL....
"""

        lvl_gen = LevelGenerator(map=map)
        lvl_gen.set_start_rect((0, 0), (18, 6))

        for _ in range(n_monster):
            lvl_gen.add_monster()

        lvl_gen.set_area_variable(
            "$boulder_area", type="fillrect", x1=1, y1=1, x2=18, y2=5
        )
        for _ in range(n_boulder):
            lvl_gen.add_object_area(
                "$boulder_area", name="boulder", symbol="`"
            )

        lvl_gen.add_goal_pos((24, 2))

        super().__init__(*args, des_file=lvl_gen.get_des(), **kwargs)


class MiniHackRiverMonster(MiniHackRiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, n_monster=5, **kwargs)


class MiniHackRiverLava(MiniHackRiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, lava=True, **kwargs)


class MiniHackRiverMonsterLava(MiniHackRiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, lava=True, n_monster=5, **kwargs)


class MiniHackRiverNarrow(MiniHackRiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, narrow=True, **kwargs)


register(
    id="MiniHack-River-v0",
    entry_point="minihack.envs.river:MiniHackRiver",
)

register(
    id="MiniHack-River-Monster-v0",
    entry_point="minihack.envs.river:MiniHackRiverMonster",
)

register(
    id="MiniHack-River-Lava-v0",
    entry_point="minihack.envs.river:MiniHackRiverLava",
)

register(
    id="MiniHack-River-MonsterLava-v0",
    entry_point="minihack.envs.river:MiniHackRiverMonsterLava",
)

register(
    id="MiniHack-River-Narrow-v0",
    entry_point="minihack.envs.river:MiniHackRiverNarrow",
)



# Copyright (c) Facebook, Inc. and its affiliates.
import os
import random

import numpy as np
import pkg_resources
from nle import nethack
from minihack.envs import register
from minihack import LevelGenerator, MiniHackNavigation

LEVELS_PATH = os.path.join(
    pkg_resources.resource_filename("minihack", "dat"),
    "boxoban-levels-master",
)
# The agent can only move towards 4 cardinal directions (instead of default 8)
MOVE_ACTIONS = tuple(nethack.CompassCardinalDirection)


def load_boxoban_levels(cur_levels_path):
    levels = []
    for file in os.listdir(cur_levels_path):
        if file.endswith(".txt"):
            with open(os.path.join(cur_levels_path, file)) as f:
                cur_lines = f.readlines()
            cur_level = []
            for el in cur_lines:
                if el != "\n":
                    cur_level.append(el)
                else:
                    # 0th element is a level number, we don't need it
                    levels.append("".join(cur_level[1:]))
                    cur_level = []
    return levels


class BoxoHack(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        kwargs["actions"] = kwargs.pop("actions", MOVE_ACTIONS)
        level_set = kwargs.pop("level_set", "unfiltered")
        level_mode = kwargs.pop("level_mode", "train")

        cur_levels_path = os.path.join(LEVELS_PATH, level_set, level_mode)

        self._time_penalty = kwargs.pop("penalty_time", 0)
        self._flags = tuple(kwargs.pop("flags", []))
        try:
            self._levels = load_boxoban_levels(cur_levels_path)
        except FileNotFoundError:
            raise ModuleNotFoundError(
                "To use Boxoban environments, please download maps using "
                "the minihack/scripts/download_boxoban_levels.py script."
            )

        self._reward_shaping_coefficient = kwargs.pop(
            "reward_shaping_coefficient", 0
        )

        super().__init__(
            *args, des_file=self.get_lvl_gen().get_des(), **kwargs
        )

    def get_env_map(self, level):
        info = {"fountains": [], "boulders": []}
        level = level[:-1]
        for row in range(0, len(level)):
            for col in range(len(level[row])):
                if level[row][col] == "$":
                    info["boulders"].append((col, row))
                if level[row][col] == ".":
                    info["fountains"].append((col, row))
            if "@" in level[row]:
                py = level[row].index("@")
                level[row] = level[row].replace("@", ".")
                info["player"] = (py, row)
            level[row] = level[row].replace(" ", ".")
            level[row] = level[row].replace("#", "F")
            level[row] = level[row].replace("$", ".")
        return "\n".join(level), info

    def get_lvl_gen(self):
        level = random.choice(self._levels)
        level = level.split("\n")
        map, info = self.get_env_map(level)
        flags = list(self._flags)
        flags.append("noteleport")
        flags.append("premapped")
        lvl_gen = LevelGenerator(map=map, lit=True, flags=flags, solidfill=" ")
        for b in info["boulders"]:
            lvl_gen.add_boulder(b)
        for f in info["fountains"]:
            lvl_gen.add_fountain(f)
        lvl_gen.set_start_pos(info["player"])
        return lvl_gen

    def reset(self, wizkit_items=None):
        self.update(self.get_lvl_gen().get_des())
        initial_obs = super().reset(wizkit_items=wizkit_items)
        self._goal_pos_set = self._object_positions(self.last_observation, "{")
        return initial_obs

    def _is_episode_end(self, observation):
        # If every boulder is on a fountain, we're done
        if self._goal_pos_set == self._object_positions(observation, "`"):
            return self.StepStatus.TASK_SUCCESSFUL
        else:
            return self.StepStatus.RUNNING

    def _reward_fn(self, last_observation, action, observation, end_status):
        if end_status == self.StepStatus.TASK_SUCCESSFUL:
            return 1
        elif end_status != self.StepStatus.RUNNING:
            return 0
        return (
            self._time_penalty
            + (
                self._count_boulders_on_fountains(observation)
                - self._count_boulders_on_fountains(last_observation)
            )
            * self._reward_shaping_coefficient
        )

    def _count_boulders_on_fountains(self, observation):
        return len(
            self._goal_pos_set.intersection(
                self._object_positions(observation, "`")
            )
        )

    def _object_positions(self, observation, object_char):
        char_obs = observation[self._original_observation_keys.index("chars")]
        return set(
            (x, y) for x, y in zip(*np.where(char_obs == ord(object_char)))
        )


class MiniHackBoxobanUnfiltered(BoxoHack):
    def __init__(self, *args, **kwargs):
        kwargs["level_set"] = "unfiltered"
        kwargs["level_mode"] = "train"
        kwargs["reward_shaping_coefficient"] = 0.1
        kwargs["penalty_time"] = -0.001
        super().__init__(*args, **kwargs)


class MiniHackBoxobanMedium(BoxoHack):
    def __init__(self, *args, **kwargs):
        kwargs["level_set"] = "medium"
        kwargs["level_mode"] = "train"
        kwargs["reward_shaping_coefficient"] = 0.1
        kwargs["penalty_time"] = -0.001
        super().__init__(*args, **kwargs)


class MiniHackBoxobanHard(BoxoHack):
    def __init__(self, *args, **kwargs):
        kwargs["level_set"] = "hard"
        kwargs["level_mode"] = ""
        kwargs["reward_shaping_coefficient"] = 0.1
        kwargs["penalty_time"] = -0.001
        super().__init__(*args, **kwargs)


register(
    id="MiniHack-Boxoban-Unfiltered-v0",
    entry_point="minihack.envs.boxohack:MiniHackBoxobanUnfiltered",
)
register(
    id="MiniHack-Boxoban-Medium-v0",
    entry_point="minihack.envs.boxohack:MiniHackBoxobanMedium",
)
register(
    id="MiniHack-Boxoban-Hard-v0",
    entry_point="minihack.envs.boxohack:MiniHackBoxobanHard",
)



# This file taken from
#     https://github.com/deepmind/scalable_agent/blob/
#         d24bd74bd53d454b7222b7f0bea57a358e4ca33e/vtrace_test.py
# and modified.

# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for V-trace.

For details and theory see:

"IMPALA: Scalable Distributed Deep-RL with
Importance Weighted Actor-Learner Architectures"
by Espeholt, Soyer, Munos et al.
"""

import unittest

import numpy as np
import torch
import vtrace


def _shaped_arange(*shape):
    """Runs np.arange, converts to float and reshapes."""
    return np.arange(np.prod(shape), dtype=np.float32).reshape(*shape)


def _softmax(logits):
    """Applies softmax non-linearity on inputs."""
    return np.exp(logits) / np.sum(np.exp(logits), axis=-1, keepdims=True)


def _ground_truth_calculation(
    discounts,
    log_rhos,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold,
    clip_pg_rho_threshold,
):
    """Calculates the ground truth for V-trace in Python/Numpy."""
    vs = []
    seq_len = len(discounts)
    rhos = np.exp(log_rhos)
    cs = np.minimum(rhos, 1.0)
    clipped_rhos = rhos
    if clip_rho_threshold:
        clipped_rhos = np.minimum(rhos, clip_rho_threshold)
    clipped_pg_rhos = rhos
    if clip_pg_rho_threshold:
        clipped_pg_rhos = np.minimum(rhos, clip_pg_rho_threshold)

    # This is a very inefficient way to calculate the V-trace ground truth.
    # We calculate it this way because it is close to the mathematical notation
    # of V-trace.
    # v_s = V(x_s)
    #             + \sum^{T-1}_{t=s} \gamma^{t-s}
    #                 * \prod_{i=s}^{t-1} c_i
    #                 * \rho_t (r_t + \gamma V(x_{t+1}) - V(x_t))
    # Note that when we take the product over c_i, we write `s:t` as the
    # notation of the paper is inclusive of the `t-1`, but Python is exclusive.
    # Also note that np.prod([]) == 1.
    values_t_plus_1 = np.concatenate(
        [values, bootstrap_value[None, :]], axis=0
    )
    for s in range(seq_len):
        v_s = np.copy(values[s])  # Very important copy.
        for t in range(s, seq_len):
            v_s += (
                np.prod(discounts[s:t], axis=0)
                * np.prod(cs[s:t], axis=0)
                * clipped_rhos[t]
                * (
                    rewards[t]
                    + discounts[t] * values_t_plus_1[t + 1]
                    - values[t]
                )
            )
        vs.append(v_s)
    vs = np.stack(vs, axis=0)
    pg_advantages = clipped_pg_rhos * (
        rewards
        + discounts
        * np.concatenate([vs[1:], bootstrap_value[None, :]], axis=0)
        - values
    )

    return vtrace.VTraceReturns(vs=vs, pg_advantages=pg_advantages)


def assert_allclose(actual, desired):
    return np.testing.assert_allclose(actual, desired, rtol=1e-06, atol=1e-05)


class ActionLogProbsTest(unittest.TestCase):
    def test_action_log_probs(self, batch_size=2):
        seq_len = 7
        num_actions = 3

        policy_logits = _shaped_arange(seq_len, batch_size, num_actions) + 10
        actions = np.random.randint(
            0, num_actions, size=(seq_len, batch_size), dtype=np.int64
        )

        action_log_probs_tensor = vtrace.action_log_probs(
            torch.from_numpy(policy_logits), torch.from_numpy(actions)
        )

        # Ground Truth
        # Using broadcasting to create a mask that indexes action logits
        action_index_mask = actions[..., None] == np.arange(num_actions)

        def index_with_mask(array, mask):
            return array[mask].reshape(*array.shape[:-1])

        # Note: Normally log(softmax) is not a good idea because it's not
        # numerically stable. However, in this test we have well-behaved values.
        ground_truth_v = index_with_mask(
            np.log(_softmax(policy_logits)), action_index_mask
        )

        assert_allclose(ground_truth_v, action_log_probs_tensor)

    def test_action_log_probs_batch_1(self):
        self.test_action_log_probs(1)


class VtraceTest(unittest.TestCase):
    def test_vtrace(self, batch_size=5):
        """Tests V-trace against ground truth data calculated in python."""
        seq_len = 5

        # Create log_rhos such that rho will span from near-zero to above the
        # clipping thresholds. In particular, calculate log_rhos in [-2.5, 2.5),
        # so that rho is in approx [0.08, 12.2).
        log_rhos = _shaped_arange(seq_len, batch_size) / (batch_size * seq_len)
        log_rhos = 5 * (log_rhos - 0.5)  # [0.0, 1.0) -> [-2.5, 2.5).
        values = {
            "log_rhos": log_rhos,
            # T, B where B_i: [0.9 / (i+1)] * T
            "discounts": np.array(
                [
                    [0.9 / (b + 1) for b in range(batch_size)]
                    for _ in range(seq_len)
                ],
                dtype=np.float32,
            ),
            "rewards": _shaped_arange(seq_len, batch_size),
            "values": _shaped_arange(seq_len, batch_size) / batch_size,
            "bootstrap_value": _shaped_arange(batch_size) + 1.0,
            "clip_rho_threshold": 3.7,
            "clip_pg_rho_threshold": 2.2,
        }

        ground_truth = _ground_truth_calculation(**values)

        values = {key: torch.tensor(value) for key, value in values.items()}
        output = vtrace.from_importance_weights(**values)

        for a, b in zip(ground_truth, output):
            assert_allclose(a, b)

    def test_vtrace_batch_1(self):
        self.test_vtrace(1)

    def test_vtrace_from_logits(self, batch_size=2):
        """Tests V-trace calculated from logits."""
        seq_len = 5
        num_actions = 3
        clip_rho_threshold = None  # No clipping.
        clip_pg_rho_threshold = None  # No clipping.

        values = {
            "behavior_policy_logits": _shaped_arange(
                seq_len, batch_size, num_actions
            ),
            "target_policy_logits": _shaped_arange(
                seq_len, batch_size, num_actions
            ),
            "actions": np.random.randint(
                0, num_actions - 1, size=(seq_len, batch_size)
            ),
            "discounts": np.array(  # T, B where B_i: [0.9 / (i+1)] * T
                [
                    [0.9 / (b + 1) for b in range(batch_size)]
                    for _ in range(seq_len)
                ],
                dtype=np.float32,
            ),
            "rewards": _shaped_arange(seq_len, batch_size),
            "values": _shaped_arange(seq_len, batch_size) / batch_size,
            "bootstrap_value": _shaped_arange(batch_size) + 1.0,  # B
        }
        values = {k: torch.from_numpy(v) for k, v in values.items()}

        from_logits_output = vtrace.from_logits(
            clip_rho_threshold=clip_rho_threshold,
            clip_pg_rho_threshold=clip_pg_rho_threshold,
            **values,
        )

        target_log_probs = vtrace.action_log_probs(
            values["target_policy_logits"], values["actions"]
        )
        behavior_log_probs = vtrace.action_log_probs(
            values["behavior_policy_logits"], values["actions"]
        )
        log_rhos = target_log_probs - behavior_log_probs

        # Calculate V-trace using the ground truth logits.
        from_iw = vtrace.from_importance_weights(
            log_rhos=log_rhos,
            discounts=values["discounts"],
            rewards=values["rewards"],
            values=values["values"],
            bootstrap_value=values["bootstrap_value"],
            clip_rho_threshold=clip_rho_threshold,
            clip_pg_rho_threshold=clip_pg_rho_threshold,
        )

        assert_allclose(from_iw.vs, from_logits_output.vs)
        assert_allclose(
            from_iw.pg_advantages, from_logits_output.pg_advantages
        )
        assert_allclose(
            behavior_log_probs, from_logits_output.behavior_action_log_probs
        )
        assert_allclose(
            target_log_probs, from_logits_output.target_action_log_probs
        )
        assert_allclose(log_rhos, from_logits_output.log_rhos)

    def test_vtrace_from_logits_batch_1(self):
        self.test_vtrace_from_logits(1)

    def test_higher_rank_inputs_for_importance_weights(self):
        """Checks support for additional dimensions in inputs."""
        T = 3  # pylint: disable=invalid-name
        B = 2  # pylint: disable=invalid-name
        values = {
            "log_rhos": torch.zeros(T, B, 1),
            "discounts": torch.zeros(T, B, 1),
            "rewards": torch.zeros(T, B, 42),
            "values": torch.zeros(T, B, 42),
            "bootstrap_value": torch.zeros(B, 42),
        }
        output = vtrace.from_importance_weights(**values)
        self.assertSequenceEqual(output.vs.shape, (T, B, 42))

    def test_inconsistent_rank_inputs_for_importance_weights(self):
        """Test one of many possible errors in shape of inputs."""
        T = 3  # pylint: disable=invalid-name
        B = 2  # pylint: disable=invalid-name

        values = {
            "log_rhos": torch.zeros(T, B, 1),
            "discounts": torch.zeros(T, B, 1),
            "rewards": torch.zeros(T, B, 42),
            "values": torch.zeros(T, B, 42),
            # Should be [B, 42].
            "bootstrap_value": torch.zeros(B),
        }

        with self.assertRaisesRegex(
            RuntimeError, "same number of dimensions: got 3 and 2"
        ):
            vtrace.from_importance_weights(**values)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation
from minihack import LevelGenerator
from minihack.envs import register


class MiniHackRoom(MiniHackNavigation):
    """Environment for "empty" task."""

    def __init__(
        self,
        *args,
        size=5,
        random=True,
        n_monster=0,
        n_trap=0,
        lit=True,
        **kwargs
    ):
        kwargs["max_episode_steps"] = kwargs.pop(
            "max_episode_steps", size * 20
        )

        lvl_gen = LevelGenerator(w=size, h=size, lit=lit)
        if random:
            lvl_gen.add_goal_pos()
        else:
            lvl_gen.add_goal_pos((size - 1, size - 1))
            lvl_gen.set_start_pos((0, 0))

        for _ in range(n_monster):
            lvl_gen.add_monster()

        for _ in range(n_trap):
            lvl_gen.add_trap()

        super().__init__(*args, des_file=lvl_gen.get_des(), **kwargs)


class MiniHackRoom5x5(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            size=5,
            random=False,
            **kwargs,
        )


class MiniHackRoom5x5Random(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=5, random=True, **kwargs)


class MiniHackRoom5x5Dark(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=5, random=True, lit=False, **kwargs)


class MiniHackRoom5x5Monster(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=5, random=True, n_monster=1, **kwargs)


class MiniHackRoom5x5Trap(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=5, random=True, n_trap=1, **kwargs)


class MiniHackRoom5x5Ultimate(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            size=5,
            random=True,
            lit=False,
            n_monster=1,
            n_trap=1,
            **kwargs,
        )


register(
    id="MiniHack-Room-5x5-v0",
    entry_point="minihack.envs.room:MiniHackRoom5x5",
)
register(
    id="MiniHack-Room-Random-5x5-v0",
    entry_point="minihack.envs.room:MiniHackRoom5x5Random",
)
register(
    id="MiniHack-Room-Dark-5x5-v0",
    entry_point="minihack.envs.room:MiniHackRoom5x5Dark",
)
register(
    id="MiniHack-Room-Monster-5x5-v0",
    entry_point="minihack.envs.room:MiniHackRoom5x5Monster",
)
register(
    id="MiniHack-Room-Trap-5x5-v0",
    entry_point="minihack.envs.room:MiniHackRoom5x5Trap",
)
register(
    id="MiniHack-Room-Ultimate-5x5-v0",
    entry_point="minihack.envs.room:MiniHackRoom5x5Ultimate",
)


class MiniHackRoom15x15(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=15, random=False, **kwargs)


class MiniHackRoom15x15Random(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=15, random=True, **kwargs)


class MiniHackRoom15x15Dark(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=15, random=True, lit=False, **kwargs)


class MiniHackRoom15x15Monster(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, size=15, random=True, n_monster=3, **kwargs)


class MiniHackRoom15x15Trap(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, size=15, random=True, n_monster=0, n_trap=15, **kwargs
        )


class MiniHackRoom15x15Ultimate(MiniHackRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            size=15,
            random=True,
            lit=False,
            n_monster=3,
            n_trap=15,
            **kwargs,
        )


register(
    id="MiniHack-Room-15x15-v0",
    entry_point="minihack.envs.room:MiniHackRoom15x15",
)
register(
    id="MiniHack-Room-Random-15x15-v0",
    entry_point="minihack.envs.room:MiniHackRoom15x15Random",
)
register(
    id="MiniHack-Room-Dark-15x15-v0",
    entry_point="minihack.envs.room:MiniHackRoom15x15Dark",
)
register(
    id="MiniHack-Room-Monster-15x15-v0",
    entry_point="minihack.envs.room:MiniHackRoom15x15Monster",
)
register(
    id="MiniHack-Room-Trap-15x15-v0",
    entry_point="minihack.envs.room:MiniHackRoom15x15Trap",
)
register(
    id="MiniHack-Room-Ultimate-15x15-v0",
    entry_point="minihack.envs.room:MiniHackRoom15x15Ultimate",
)



# Copyright (c) Facebook, Inc. and its affiliates.

import json
import os
import re
from collections import defaultdict
from functools import lru_cache
from typing import List
from urllib.parse import unquote

import pkg_resources

try:
    import inflect
    import stanza

    PREPROCESSING_ALLOWED = True
    import_error = None
except ImportError as error:  # noqa
    PREPROCESSING_ALLOWED = False
    import_error = error

DATA_DIR_PATH = pkg_resources.resource_filename("nle", "minihack/dat")

EXCEPTIONS = (
    "floor of a room",
    "agent",
    "staircase up",
)


class TextProcessor:
    """Base class for modeling relations between an object and subject."""

    def __init__(self):
        # Will only do it the first time
        stanza.download("en")
        self.nlp = stanza.Pipeline(lang="en", processors="tokenize,mwt,pos")
        self.inflect = inflect.engine()

    @lru_cache(maxsize=None)
    def preprocess(self, input_str: str) -> str:
        # Removes the brackets and non-letter charachters
        text = re.sub(r"\([^)]*\)", "", input_str)
        pattern = re.compile(r"[^a-zA-Z]+")
        text = pattern.sub(" ", text)
        # Remove trailing whitespaces
        text = re.sub(r"\w[ ]{2,}\w", " ", text)
        return text.strip()

    @lru_cache(maxsize=None)
    def process(self, input_str: str) -> str:
        input_str = self.preprocess(input_str)

        # First find nouns in phrase
        result = self.nlp(input_str)
        nouns = [
            word.text
            for sent in result.sentences
            for word in sent.words
            if word.upos in {"NOUN", "PROPN"}
        ]
        if not nouns:
            return input_str
        # Pick last noun in input
        noun = nouns[-1]

        # Singularise the noun - returns False if the word is alread singular
        singular = self.inflect.singular_noun(noun.lower())
        if not singular:
            return nouns[-1].lower()
        else:
            return singular


class NetHackWiki:
    """A class representing Nethack Wiki Data - pages and links between them.

    Args:
        raw_wiki_file_name (str):
            The path to the raw file of NetHack wiki. The raw file can be
            downloaded using the `get_nhwiki_data.sh` script located in
            `minihack/scripts`.
        processed_wiki_file_name (str):
            The path to the processed file of NetHack wiki. The processing
            is performed in the `__init__` function of this classed.
        save_processed_json (bool):
            Whether to save the processed json file of the wiki. Only
            considered when a raw wiki file is passed. Defaults to True.
        ignore_inpage_anchors (bool):
            Whether to ingnore in-page anchors. Defaults to True.
        preprocess_input (bool):
            Whether to perform a preprocessing on wiki data. Defaults to True.
        exceptions (Tuple[str] or None):
            Name of entities in screen descriptions that are ingored. If None,
            there are no exceptions. Defaults to None.
    """

    def __init__(
        self,
        raw_wiki_file_name: str,
        processed_wiki_file_name: str,
        save_processed_json: bool = True,
        ignore_inpage_anchors: bool = True,
        preprocess_input: bool = True,
        exceptions: tuple = None,
    ) -> None:
        if os.path.isfile(processed_wiki_file_name):
            with open(processed_wiki_file_name, "r") as json_file:
                self.wiki = json.load(json_file)
        elif os.path.isfile(raw_wiki_file_name):
            raw_json = load_json(raw_wiki_file_name)
            self.wiki = process_json(
                raw_json, ignore_inpage_anchors=ignore_inpage_anchors
            )
            if save_processed_json:
                with open(processed_wiki_file_name, "w+") as json_file:
                    json.dump(self.wiki, json_file)
        else:
            raise ValueError(
                """One of `raw_wiki_file_name` or `processed_wiki_file_name`
                must be supplied as argument and be a file. Try using
                `nle/minihack/scripts/get_nhwiki_data.sh` to download the
                data."""
            )

        self.exceptions = exceptions if exceptions is not None else EXCEPTIONS
        self.preprocess_input = preprocess_input
        if preprocess_input:
            if PREPROCESSING_ALLOWED:
                self.text_processor = TextProcessor()
            else:
                print(
                    "To perform text preprocessing, `inflect` and `stanza`"
                    f"must be installed. See {import_error} for more information"
                )
                self.preprocess_input = False

    def get_page_text(self, page: str) -> str:
        """Get the text of a page.

        Args:
            page (str): The page name.
        Returns:
            str: The text of the page.
        """
        if page in self.exceptions:
            return ""
        if self.preprocess_input:
            page = self.text_processor.process(page)
        return self.wiki.get(page, {}).get("text", "")

    def get_page_data(self, page: str) -> dict:
        """Get the data of a page.

        Args:
            page (str): The page name.
        Returns:
            dict: The page data as a dict.
        """
        if page in self.exceptions:
            return {}
        if self.preprocess_input:
            page = self.text_processor.process(input)
        return self.wiki.get(page, {})


def load_json(file_name: str) -> list:
    """Load a file containing a json object per line into a list of dicts."""
    with open(file_name, "r") as json_file:
        input_json = []
        for line in json_file:
            input_json.append(json.loads(line))
    return input_json


def process_json(wiki_json: List[dict], ignore_inpage_anchors) -> dict:
    """Process a list of json pages of the wiki into one dict of all pages."""
    result: dict = {}
    redirects = {}
    result["_global_counts"] = defaultdict(int)

    def href_normalise(x: str):
        result = unquote(x.lower())
        if ignore_inpage_anchors:
            result = result.split("#")[0]
        return result.replace("_", " ")

    for page in wiki_json:
        relevant_page_info = dict(
            title=page["wikipedia_title"].lower(),
            length=len("".join(page["text"])),
            categories=page["categories"].split(","),
            raw_text="".join(page["text"]),
            text=clean_page_text(page["page_data"]),
        )
        # noqa: E731
        relevant_page_info["anchors"] = [
            dict(
                text=anchor["text"].lower(),
                page=href_normalise(anchor.get("title", anchor.get("href"))),
                start=anchor["start"],
            )
            for anchor in page["anchors"]
        ]
        redirect_anchors = [
            anchor
            for anchor in page["anchors"]
            if anchor.get("title")
            and href_normalise(anchor["href"])
            != href_normalise(anchor["title"])
        ]
        redirects.update(
            {
                href_normalise(anchor["href"]): href_normalise(anchor["title"])
                for anchor in redirect_anchors
            }
        )
        unique_anchors: dict = defaultdict(int)
        for anchor in relevant_page_info["anchors"]:
            unique_anchors[anchor["page"]] += 1
            result["_global_counts"][anchor["page"]] += 1
        relevant_page_info["unique_anchors"] = dict(unique_anchors)
        result[relevant_page_info["title"]] = relevant_page_info
    for alias, page in redirects.items():
        result[alias] = result[page]
    return result


def clean_page_text(text: List[str]) -> str:
    """Clean Markdown text to make it more passable into an NLP model.

    This is currently very basic, and more advanced parsing could be employed
    if necessary."""

    return re.sub(r"[^a-zA-Z0-9_\s\.]", "", ",".join(text))



#!/usr/bin/env python3
# This file is adapted from
# github.com/maximecb/gym-minigrid/blob/master/manual_control.py

import argparse
import numpy as np
import gym
from nle import nethack
import minihack  # noqa
from minihack.tiles.window import Window


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--env",
        help="gym environment to load",
        default="MiniHack-CorridorBattle-v0",
    )

    args = parser.parse_args()

    observation_keys = ("pixel", "message")
    env = gym.make(args.env, observation_keys=observation_keys)

    def reset():
        obs = env.reset()
        redraw(obs)

    def step(action):
        obs, reward, done, info = env.step(action)

        if done:
            print("Episode Completed!")
            reset()
        else:
            redraw(obs)

    def key_handler(event):
        if event.key == "escape":
            window.close()
            return

        if event.key == "backspace":
            reset()
            return

        if event.key.startswith("ctrl+"):
            ch = nethack.C(event.key[5])
        else:
            ch = ord(event.key)

        try:
            action = env._actions.index(ch)
            step(action)
        except (ValueError, TypeError):
            print(
                f"Selected action {event.key} is not in action list. "
                "Please try again."
            )

    window = Window("MiniHack the Planet - " + args.env)
    window.reg_key_handler(key_handler)

    def redraw(obs):
        img = obs["pixel"]
        msg = obs["message"]
        msg = msg[: np.where(msg == 0)[0][0]].tobytes().decode("utf-8")
        window.show_obs(img, msg)

    reset()

    # Blocking event loop
    window.show(block=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
#  MINIHACK_RELEASE_BUILD
#    If set, builds wheel (s)dist such as to prepare it for upload to PyPI.
#

import os
import setuptools
import subprocess


packages = [
    "minihack",
    "minihack.envs",
    "minihack.scripts",
    "minihack.tiles",
    "minihack.tests",
    "minihack.agent",
    "minihack.agent.polybeast",
    "minihack.agent.polybeast.models",
    "minihack.agent.polybeast.core",
    "minihack.agent.rllib",
    "minihack.agent.common",
    "minihack.agent.common.envs",
    "minihack.agent.common.models",
    "minihack.agent.common.util",
]

entry_points = {
    "console_scripts": [
        "mh-play = minihack.scripts.play:main",
        "mh-guiplay = minihack.scripts.play_gui:main",
        "mh-envs = minihack.scripts.env_list:main",
    ]
}

install_requires = ["numpy>=1.16", "gym"]
if not os.getenv("READTHEDOCS"):
    install_requires.append("nle>=0.8.0")

extras_deps = {
    "dev": [
        "pre-commit>=2.0.1",
        "black>=19.10b0",
        "flake8>=3.7",
        "flake8-bugbear>=20.1",
        "pytest>=5.3",
        "pytest-benchmark>=3.1.0",
        "sphinx==4.0.2",
        "sphinx-rtd-theme==1.0.0",
        "myst-parser==0.15.1",
        "nbsphinx==0.8.6",
    ],
    "polybeast": [
        "torch>=1.3.1",
        "hydra-core>=1.0.0",
        "hydra-colorlog>=1.0.0",
        "hydra-submitit-launcher>=1.1.1",
        "wandb>=0.10.31",
        "pyyaml",
    ],
    "rllib": [
        "torch>=1.3.1",
        "ray[rllib]==1.3.0",
        "ray[default]==1.3.0",
        "hydra-core>=1.0.0",
        "hydra-colorlog>=1.0.0",
        "hydra-submitit-launcher>=1.1.1",
        "wandb>=0.10.31",
    ],
    "wiki": [
        "inflect",
        "stanza",
    ],
}

extras_deps["all"] = [item for group in extras_deps.values() for item in group]


if __name__ == "__main__":
    with open("README.md") as f:
        long_description = f.read()
    cwd = os.path.dirname(os.path.abspath(__file__))
    sha = "Unknown"
    version = open("version.txt", "r").read().strip()

    try:
        sha = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        pass

    if sha != "Unknown" and not os.getenv("MINIHACK_RELEASE_BUILD"):
        version += "+" + sha[:7]
    print("Building wheel {}-{}".format("minihack", version))

    version_path = os.path.join(cwd, "minihack", "version.py")
    with open(version_path, "w") as f:
        f.write("__version__ = '{}'\n".format(version))
        f.write("git_version = {}\n".format(repr(sha)))

    setuptools.setup(
        name="minihack",
        version=version,
        description="MiniHack The Planet: "
        + "A Sandbox for Open-Ended Reinforcement Learning Research",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="The MiniHack Team",
        url="https://github.com/facebookresearch/minihack",
        license="Apache License, Version 2.0",
        entry_points=entry_points,
        packages=packages,
        install_requires=install_requires,
        extras_require=extras_deps,
        python_requires=">=3.7",
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Development Status :: 4 - Beta",
            "Operating System :: POSIX :: Linux",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Games/Entertainment",
        ],
        zip_safe=False,
        include_package_data=True,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation
from minihack.level_generator import PATH_DAT_DIR
from minihack.envs import register
from nle.nethack import Command
from nle import nethack
import os

MOVE_ACTIONS = tuple(nethack.CompassDirection)
APPLY_ACTIONS = tuple(list(MOVE_ACTIONS) + [Command.PICKUP, Command.APPLY])


class KeyRoomGenerator:
    def __init__(self, room_size, subroom_size, lit):
        des_path = os.path.join(PATH_DAT_DIR, "key_and_door_tmp.des")
        with open(des_path) as f:
            df = f.read()

        df = df.replace("RS", str(room_size))
        df = df.replace("SS", str(subroom_size))
        if not lit:
            df = df.replace("lit", str("unlit"))

        self.des_file = df

    def get_des(self):
        return self.des_file


class MiniHackKeyDoor(MiniHackNavigation):
    """Environment for "key and door" task."""

    def __init__(self, *args, des_file, **kwargs):
        kwargs["character"] = kwargs.pop("charachter", "rog-hum-cha-mal")
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        kwargs["actions"] = kwargs.pop("actions", APPLY_ACTIONS)
        kwargs["autopickup"] = kwargs.pop("autopickup", False)
        super().__init__(*args, des_file=des_file, **kwargs)

    def step(self, action: int):
        # If apply action is chosen
        if self._actions[action] == Command.APPLY:
            key_key = self.key_in_inventory("key")
            # if key is in the inventory
            if key_key is not None:
                # Check if there is a closed door nearby
                dir_key = self.get_object_direction("closed door")
                if dir_key is not None:
                    # Perform the following NetHack steps
                    self.env.step(Command.APPLY)  # press apply
                    self.env.step(ord(key_key))  # choose key from the inv
                    self.env.step(dir_key)  # select the door's direction
                    obs, done = self.env.step(ord("y"))  # press y
                    obs, done = self._perform_known_steps(
                        obs, done, exceptions=True
                    )
                    # Make sure the door is open
                    while True:
                        obs, done = self.env.step(dir_key)
                        obs, done = self._perform_known_steps(
                            obs, done, exceptions=True
                        )
                        if (
                            self.get_object_direction("closed door", obs)
                            is None
                        ):
                            break

        obs, reward, done, info = super().step(action)
        return obs, reward, done, info


class MiniHackKeyRoom(MiniHackKeyDoor):
    def __init__(self, *args, room_size, subroom_size, lit, **kwargs):
        lev_gen = KeyRoomGenerator(room_size, subroom_size, lit)
        des_file = lev_gen.get_des()
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackKeyRoom5x5Fixed(MiniHackKeyDoor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="key_and_door.des", **kwargs)


class MiniHackKeyRoom5x5(MiniHackKeyRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, room_size=5, subroom_size=2, lit=True, **kwargs
        )


class MiniHackKeyRoom5x5Dark(MiniHackKeyRoom):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, room_size=5, subroom_size=2, lit=False, **kwargs
        )


class MiniHackKeyRoom15x15(MiniHackKeyRoom):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        super().__init__(
            *args, room_size=15, subroom_size=5, lit=True, **kwargs
        )


class MiniHackKeyRoom15x15Dark(MiniHackKeyRoom):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        super().__init__(
            *args, room_size=15, subroom_size=5, lit=False, **kwargs
        )


register(
    id="MiniHack-KeyRoom-Fixed-S5-v0",
    entry_point="minihack.envs.keyroom:MiniHackKeyRoom5x5Fixed",
)
register(
    id="MiniHack-KeyRoom-S5-v0",
    entry_point="minihack.envs.keyroom:MiniHackKeyRoom5x5",
)
register(
    id="MiniHack-KeyRoom-S15-v0",
    entry_point="minihack.envs.keyroom:MiniHackKeyRoom15x15",
)
register(
    id="MiniHack-KeyRoom-Dark-S5-v0",
    entry_point="minihack.envs.keyroom:MiniHackKeyRoom5x5Dark",
)
register(
    id="MiniHack-KeyRoom-Dark-S15-v0",
    entry_point="minihack.envs.keyroom:MiniHackKeyRoom15x15Dark",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import Any, Dict, Optional, Tuple

import gym
import torch
from nle import nethack
from minihack.agent.common.models.embed import GlyphEmbedding
from minihack.agent.common.models.transformer import TransformerEncoder
from omegaconf import DictConfig
from ray.rllib.models import ModelCatalog
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.utils.annotations import override
from torch import nn
from torch.nn import functional as F

NUM_GLYPHS = nethack.MAX_GLYPH
NUM_FEATURES = nethack.BLSTATS_SHAPE[0]
PAD_CHAR = 0
NUM_CHARS = 128


class RLLibGlyphEmbedding(GlyphEmbedding):
    def glyphs_to_idgroup(self, glyphs):
        B, H, W = glyphs.shape
        ids_groups = self.id_pairs_table.index_select(
            0, glyphs.contiguous().view(-1).long()
        )
        ids = ids_groups.select(1, 0).view(B, H, W).long()
        groups = ids_groups.select(1, 1).view(B, H, W).long()
        return (ids, groups)

    def prepare_input(self, inputs):
        """Take the inputs to the network as dictionary and return a namedtuple
        of the input/index tensors to be embedded (GlyphTuple)"""
        embeddable_data = {}
        # Only flatten the data we want
        for key, value in inputs.items():
            if key in self.embeddings:
                # -- [ B x ...] -> [ B' x ... ]
                # embeddable_data[key] = torch.flatten(value, 0, 1).long()
                embeddable_data[key] = value.long()

        # add our group id and subgroup id if we want them
        if self.requires_id_pairs_table:
            ids, groups = self.glyphs_to_idgroup(inputs["glyphs"])
            embeddable_data["groups"] = groups
            embeddable_data["subgroup_ids"] = ids

        # convert embeddable_data to a named tuple
        return self.GlyphTuple(**embeddable_data)


class NetHackNet(nn.Module):
    AgentOutput = collections.namedtuple(
        "AgentOutput", "action policy_logits baseline"
    )

    def __init__(self):
        super(NetHackNet, self).__init__()

        self.register_buffer("reward_sum", torch.zeros(()))
        self.register_buffer("reward_m2", torch.zeros(()))
        self.register_buffer("reward_count", torch.zeros(()).fill_(1e-8))

    def forward(self, inputs, core_state):
        raise NotImplementedError

    def initial_state(self, batch_size=1):
        return ()

    def prepare_input(self, inputs):
        # -- [B x H x W]
        glyphs = inputs["glyphs"]

        # -- [B x F]
        features = inputs["blstats"]

        B, *_ = glyphs.shape

        return glyphs, features

    def embed_state(self, inputs):
        raise NotImplementedError

    @torch.no_grad()
    def update_running_moments(self, reward_batch):
        """Maintains a running mean of reward."""
        new_count = len(reward_batch)
        new_sum = torch.sum(reward_batch)
        new_mean = new_sum / new_count

        curr_mean = self.reward_sum / self.reward_count
        new_m2 = torch.sum((reward_batch - new_mean) ** 2) + (
            (self.reward_count * new_count)
            / (self.reward_count + new_count)
            * (new_mean - curr_mean) ** 2
        )

        self.reward_count += new_count
        self.reward_sum += new_sum
        self.reward_m2 += new_m2

    @torch.no_grad()
    def get_running_std(self):
        """Returns standard deviation of the running mean of the reward."""
        return torch.sqrt(self.reward_m2 / self.reward_count)


class Crop(nn.Module):
    def __init__(self, height, width, height_target, width_target):
        super(Crop, self).__init__()
        self.width = width
        self.height = height
        self.width_target = width_target
        self.height_target = height_target

        width_grid = self._step_to_range(
            2 / (self.width - 1), self.width_target
        )[None, :].expand(self.height_target, -1)
        height_grid = self._step_to_range(
            2 / (self.height - 1), height_target
        )[:, None].expand(-1, self.width_target)

        # "clone" necessary, https://github.com/pytorch/pytorch/issues/34880
        self.register_buffer("width_grid", width_grid.clone())
        self.register_buffer("height_grid", height_grid.clone())

    def _step_to_range(self, step, num_steps):
        return torch.tensor(
            [step * (i - num_steps // 2) for i in range(num_steps)]
        )

    def forward(self, inputs, coordinates):
        """Calculates centered crop around given x,y coordinates.

        Args:
           inputs [B x H x W] or [B x H x W x C]
           coordinates [B x 2] x,y coordinates

        Returns:
           [B x H' x W'] inputs cropped and centered around x,y coordinates.
        """
        assert inputs.shape[1] == self.height, "expected %d but found %d" % (
            self.height,
            inputs.shape[1],
        )
        assert inputs.shape[2] == self.width, "expected %d but found %d" % (
            self.width,
            inputs.shape[2],
        )

        permute_results = False
        if inputs.dim() == 3:
            inputs = inputs.unsqueeze(1)
        else:
            permute_results = True
            inputs = inputs.permute(0, 2, 3, 1)
        inputs = inputs.float()

        x = coordinates[:, 0]
        y = coordinates[:, 1]

        x_shift = 2 / (self.width - 1) * (x.float() - self.width // 2)
        y_shift = 2 / (self.height - 1) * (y.float() - self.height // 2)

        grid = torch.stack(
            [
                self.width_grid[None, :, :] + x_shift[:, None, None],
                self.height_grid[None, :, :] + y_shift[:, None, None],
            ],
            dim=3,
        )

        crop = (
            torch.round(F.grid_sample(inputs, grid, align_corners=True))
            .squeeze(1)
            .long()
        )

        if permute_results:
            # [B x C x H x W] -> [B x H x W x C]
            crop = crop.permute(0, 2, 3, 1)

        return crop


class Flatten(nn.Module):
    def forward(self, input):
        return input.view(input.size(0), -1)


class BaseNet(NetHackNet):
    def __init__(self, processed_observation_shape, flags: DictConfig):
        super(BaseNet, self).__init__()

        self.observation_space = processed_observation_shape.original_space

        self.H = self.observation_space["glyphs"].shape[0]
        self.W = self.observation_space["glyphs"].shape[1]

        self.k_dim = flags.embedding_dim
        self.h_dim = flags.hidden_dim

        self.crop_model = flags.crop_model
        self.crop_dim = flags.crop_dim

        self.num_features = NUM_FEATURES

        self.crop = Crop(self.H, self.W, self.crop_dim, self.crop_dim)

        self.glyph_type = flags.glyph_type
        self.glyph_embedding = RLLibGlyphEmbedding(
            flags.glyph_type,
            flags.embedding_dim,
            None,
            flags.use_index_select,
        )

        K = flags.embedding_dim  # number of input filters
        F = 3  # filter dimensions
        S = 1  # stride
        P = 1  # padding
        M = 16  # number of intermediate filters
        self.Y = 8  # number of output filters
        L = flags.layers  # number of convnet layers

        in_channels = [K] + [M] * (L - 1)
        out_channels = [M] * (L - 1) + [self.Y]

        def interleave(xs, ys):
            return [val for pair in zip(xs, ys) for val in pair]

        conv_extract = [
            nn.Conv2d(
                in_channels=in_channels[i],
                out_channels=out_channels[i],
                kernel_size=(F, F),
                stride=S,
                padding=P,
            )
            for i in range(L)
        ]

        self.extract_representation = nn.Sequential(
            *interleave(conv_extract, [nn.ELU()] * len(conv_extract))
        )

        if self.crop_model == "transformer":
            self.extract_crop_representation = TransformerEncoder(
                K,
                N=L,
                heads=8,
                height=self.crop_dim,
                width=self.crop_dim,
                device=None,
            )
        elif self.crop_model == "cnn":
            conv_extract_crop = [
                nn.Conv2d(
                    in_channels=in_channels[i],
                    out_channels=out_channels[i],
                    kernel_size=(F, F),
                    stride=S,
                    padding=P,
                )
                for i in range(L)
            ]

            self.extract_crop_representation = nn.Sequential(
                *interleave(conv_extract_crop, [nn.ELU()] * len(conv_extract))
            )

        # MESSAGING MODEL
        if "msg" not in flags:
            self.msg_model = "none"
        else:
            self.msg_model = flags.msg.model
            self.msg_hdim = flags.msg.hidden_dim
            self.msg_edim = flags.msg.embedding_dim
        if self.msg_model in ("gru", "lstm", "lt_cnn"):
            # character-based embeddings
            self.char_lt = nn.Embedding(
                NUM_CHARS, self.msg_edim, padding_idx=PAD_CHAR
            )
        else:
            # forward will set up one-hot inputs for the cnn, no lt needed
            pass

        if self.msg_model.endswith("cnn"):
            # from Zhang et al, 2016
            # Character-level Convolutional Networks for Text Classification
            # https://arxiv.org/abs/1509.01626
            if self.msg_model == "cnn":
                # inputs will be one-hot vectors, as done in paper
                self.conv1 = nn.Conv1d(NUM_CHARS, self.msg_hdim, kernel_size=7)
            elif self.msg_model == "lt_cnn":
                # replace one-hot inputs with learned embeddings
                self.conv1 = nn.Conv1d(
                    self.msg_edim, self.msg_hdim, kernel_size=7
                )
            else:
                raise NotImplementedError("msg.model == %s", flags.msg.model)

            # remaining convolutions, relus, pools, and a small FC network
            self.conv2_6_fc = nn.Sequential(
                nn.ReLU(),
                nn.MaxPool1d(kernel_size=3, stride=3),
                # conv2
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=7),
                nn.ReLU(),
                nn.MaxPool1d(kernel_size=3, stride=3),
                # conv3
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                # conv4
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                # conv5
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                # conv6
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                nn.MaxPool1d(kernel_size=3, stride=3),
                # fc receives -- [ B x h_dim x 5 ]
                Flatten(),
                nn.Linear(5 * self.msg_hdim, 2 * self.msg_hdim),
                nn.ReLU(),
                nn.Linear(2 * self.msg_hdim, self.msg_hdim),
            )  # final output -- [ B x h_dim x 5 ]
        elif self.msg_model in ("gru", "lstm"):

            def rnn(flag):
                return nn.LSTM if flag == "lstm" else nn.GRU

            self.char_rnn = rnn(self.msg_model)(
                self.msg_edim,
                self.msg_hdim // 2,
                batch_first=True,
                bidirectional=True,
            )
        elif self.msg_model != "none":
            raise NotImplementedError("msg.model == %s", flags.msg.model)

        self.embed_features = nn.Sequential(
            nn.Linear(self.num_features, self.k_dim),
            nn.ReLU(),
            nn.Linear(self.k_dim, self.k_dim),
            nn.ReLU(),
        )

        self.equalize_input_dim = flags.equalize_input_dim
        if not self.equalize_input_dim:
            # just added up the output dimensions of the input featurizers
            # feature / status dim
            out_dim = self.k_dim
            # CNN over full glyph map
            out_dim += self.H * self.W * self.Y
            if self.crop_model == "transformer":
                out_dim += self.crop_dim ** 2 * K
            elif self.crop_model == "cnn":
                out_dim += self.crop_dim ** 2 * self.Y
            # messaging model
            if self.msg_model != "none":
                out_dim += self.msg_hdim
        else:
            # otherwise, project them all to h_dim
            NUM_INPUTS = 4 if self.msg_model != "none" else 3
            project_hdim = flags.equalize_factor * self.h_dim
            out_dim = project_hdim * NUM_INPUTS

            # set up linear layers for projections
            self.project_feature_dim = nn.Linear(self.k_dim, project_hdim)
            self.project_glyph_dim = nn.Linear(
                self.H * self.W * self.Y, project_hdim
            )
            c__2 = self.crop_dim ** 2
            if self.crop_model == "transformer":
                self.project_crop_dim = nn.Linear(c__2 * K, project_hdim)
            elif self.crop_model == "cnn":
                self.project_crop_dim = nn.Linear(c__2 * self.Y, project_hdim)
            if self.msg_model != "none":
                self.project_msg_dim = nn.Linear(self.msg_hdim, project_hdim)

        self.fc = nn.Sequential(
            nn.Linear(out_dim, self.h_dim),
            nn.ReLU(),
            nn.Linear(self.h_dim, self.h_dim),
            nn.ReLU(),
        )

    def prepare_input(self, inputs):
        # -- [B x H x W]
        B, H, W = inputs["glyphs"].shape

        # take our chosen glyphs and merge the time and batch

        glyphs = self.glyph_embedding.prepare_input(inputs)

        # -- [B x F]
        features = inputs["blstats"]

        return glyphs, features

    def forward(self, inputs):
        B, *_ = inputs["glyphs"].shape

        glyphs, features = self.prepare_input(inputs)

        # -- [B x 2] x,y coordinates
        coordinates = features[:, :2]

        # -- [B x K]
        features_emb = self.embed_features(features)
        if self.equalize_input_dim:
            features_emb = self.project_feature_dim(features_emb)

        assert features_emb.shape[0] == B

        reps = [features_emb]  # either k_dim or project_hdim

        # -- [B x H' x W']
        crop = self.glyph_embedding.GlyphTuple(
            *[self.crop(g, coordinates) for g in glyphs]
        )
        # -- [B x H' x W' x K]
        crop_emb = self.glyph_embedding(crop)

        if self.crop_model == "transformer":
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb, mask=None)
        elif self.crop_model == "cnn":
            # -- [B x K x W' x H']
            crop_emb = crop_emb.transpose(1, 3)
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb)

        # -- [B x K']
        crop_rep = crop_rep.view(B, -1)
        if self.equalize_input_dim:
            crop_rep = self.project_crop_dim(crop_rep)
        assert crop_rep.shape[0] == B

        reps.append(crop_rep)  # either k_dim or project_hdim

        # -- [B x H x W x K]
        glyphs_emb = self.glyph_embedding(glyphs)
        # glyphs_emb = self.embed(glyphs)
        # -- [B x K x W x H]
        glyphs_emb = glyphs_emb.transpose(1, 3)
        # -- [B x W x H x K]
        glyphs_rep = self.extract_representation(glyphs_emb)

        # -- [B x K']
        glyphs_rep = glyphs_rep.view(B, -1)
        # -- [B x K']
        if self.equalize_input_dim:
            glyphs_rep = self.project_glyph_dim(glyphs_rep)

        assert glyphs_rep.shape[0] == B

        # -- [B x K'']
        reps.append(glyphs_rep)

        # MESSAGING MODEL
        if self.msg_model != "none":
            messages = inputs["message"].long()
            if self.msg_model == "cnn":
                # convert messages to one-hot, [B x 96 x 256]
                one_hot = F.one_hot(messages, num_classes=NUM_CHARS).transpose(
                    1, 2
                )
                char_rep = self.conv2_6_fc(self.conv1(one_hot.float()))
            elif self.msg_model == "lt_cnn":
                # [B x E x 256 ]
                char_emb = self.char_lt(messages).transpose(1, 2)
                char_rep = self.conv2_6_fc(self.conv1(char_emb))
            else:  # lstm, gru
                char_emb = self.char_lt(messages)
                output = self.char_rnn(char_emb)[0]
                fwd_rep = output[:, -1, : self.h_dim // 2]
                bwd_rep = output[:, 0, self.h_dim // 2 :]
                char_rep = torch.cat([fwd_rep, bwd_rep], dim=1)

            if self.equalize_input_dim:
                char_rep = self.project_msg_dim(char_rep)
            reps.append(char_rep)

        st = torch.cat(reps, dim=1)

        # -- [B x K]
        st = self.fc(st)

        return st


class RLLibNLENetwork(TorchModelV2, nn.Module):
    def __init__(
        self,
        observation_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: Optional[int],
        model_config: dict,
        name: str,
        **kwargs: Any,
    ):
        TorchModelV2.__init__(
            self,
            observation_space,
            action_space,
            num_outputs,
            model_config,
            name,
        )
        nn.Module.__init__(self)

        flags = model_config["custom_model_config"]["flags"]
        self.num_outputs = num_outputs or flags.hidden_dim

        self.base = BaseNet(observation_space, flags)  # device is sorted later

    @override(TorchModelV2)
    def forward(self, x: Dict[str, Any], *_: Any) -> Tuple[torch.Tensor, list]:
        return self.base(x["obs"]), []


ModelCatalog.register_custom_model("rllib_nle_model", RLLibNLENetwork)



# This file is adapted from
# https://github.com/maximecb/gym-minigrid/blob/master/gym_minigrid/window.py


import sys

# Only ask users to install matplotlib if they actually need it
try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    print(
        "To display the environment in a window, please install matplotlib, eg:"
    )
    print("pip3 install --user matplotlib")
    sys.exit(-1)


class Window:
    """
    Window to draw a gridworld instance using Matplotlib
    """

    def __init__(self, title):
        self.fig = None

        self.imshow_obj = None

        # Create the figure and axes
        self.fig, self.ax = plt.subplots()

        # Show the env name in the window title
        self.fig.canvas.set_window_title(title)

        # Turn off x/y axis numbering/ticks
        self.ax.xaxis.set_ticks_position("none")
        self.ax.yaxis.set_ticks_position("none")
        _ = self.ax.set_xticklabels([])
        _ = self.ax.set_yticklabels([])

        # Flag indicating the window was closed
        self.closed = False
        self.title = ""

        def close_handler(evt):
            self.closed = True

        self.fig.canvas.mpl_connect("close_event", close_handler)

        # Remove all keymaps
        for key in plt.rcParams.keys():
            if "keymap" in key:
                plt.rcParams[key].clear()

    def show_obs(self, img, msg=""):
        """
        Show an rbg tiles of the observation and in-game message
        """

        # Show the first image of the environment
        if self.imshow_obj is None:
            self.imshow_obj = self.ax.imshow(img, interpolation="bilinear")

        self.imshow_obj.set_data(img)
        self.fig.canvas.draw()

        if msg != self.title:
            self.title = msg
            plt.title(msg, loc="left")

        # Let matplotlib process UI events
        # This is needed for interactive mode to work properly
        plt.pause(0.001)

    def set_caption(self, text):
        """
        Set/update the caption text below the image
        """

        plt.xlabel(text)

    def reg_key_handler(self, key_handler):
        """
        Register a keyboard event handler
        """

        # Keyboard handler
        self.fig.canvas.mpl_connect("key_press_event", key_handler)

    def show(self, block=True):
        """
        Show the window, and start an event loop
        """

        # If not blocking, trigger interactive mode
        if not block:
            plt.ion()

        # Show the plot
        # In non-interative mode, this enters the matplotlib event loop
        # In interactive mode, this call does not block
        plt.show()

    def close(self):
        """
        Close the window
        """

        plt.close()
        self.closed = True



# Copyright (c) Facebook, Inc. and its affiliates.

from minihack.level_generator import LevelGenerator
from minihack.reward_manager import RewardManager
from minihack.base import MiniHack
from minihack.navigation import MiniHackNavigation
from minihack.skills import MiniHackSkill
from minihack.wiki import NetHackWiki

import minihack.navigation
import minihack.skills
import minihack.envs.room
import minihack.envs.corridor
import minihack.envs.keyroom
import minihack.envs.mazewalk
import minihack.envs.fightcorridor
import minihack.envs.minigrid
import minihack.envs.memento
import minihack.envs.boxohack
import minihack.envs.river
import minihack.envs.hidenseek
import minihack.envs.lab
import minihack.envs.exploremaze
import minihack.envs.skills_simple
import minihack.envs.skills_wod
import minihack.envs.skills_levitate
import minihack.envs.skills_freeze
import minihack.envs.skills_lava
import minihack.envs.skills_quest

__all__ = [
    "MiniHack",
    "MiniHackNavigation",
    "MiniHackSkill",
    "LevelGenerator",
    "RewardManager",
    "NetHackWiki",
]

from .version import __version__



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackSkill
from minihack.envs import register


class MiniHackLCLevitatePotionPickup(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
-------------
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
-------------
ENDMAP
REGION:(0,0,12,6),lit,"ordinary"
$left_bank = selection:fillrect (1,1,5,5)
$right_bank = selection:fillrect (7,1,11,5)
OBJECT:('!',"levitation"),rndcoord($left_bank),blessed
BRANCH:(1,1,5,5),(0,0,0,0)
STAIR:rndcoord($right_bank),down
"""
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLCLevitatePotionInv(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        kwargs["autopickup"] = kwargs.pop("autopickup", True)
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
-------------
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
-------------
ENDMAP
REGION:(0,0,12,6),lit,"ordinary"
$right_bank = selection:fillrect (7,1,11,5)
OBJECT:('!',"levitation"),(2,2),blessed
BRANCH:(2,2,2,2),(0,0,0,0)
STAIR:rndcoord($right_bank),down
"""
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLCLevitateRingPickup(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
-------------
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
-------------
ENDMAP
REGION:(0,0,12,6),lit,"ordinary"
$left_bank = selection:fillrect (1,1,5,5)
$right_bank = selection:fillrect (7,1,11,5)
OBJECT:('=',"levitation"),rndcoord($left_bank),blessed
BRANCH:(1,1,5,5),(0,0,0,0)
STAIR:rndcoord($right_bank),down
"""
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLCLevitateRingInv(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        kwargs["autopickup"] = kwargs.pop("autopickup", True)
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
-------------
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
-------------
ENDMAP
REGION:(0,0,12,6),lit,"ordinary"
$right_bank = selection:fillrect (7,1,11,5)
OBJECT:('=',"levitation"),(2,2),blessed
BRANCH:(2,2,2,2),(0,0,0,0)
STAIR:rndcoord($right_bank),down
"""
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLCLevitate(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
-------------
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
-------------
ENDMAP
REGION:(0,0,12,6),lit,"ordinary"
$left_bank = selection:fillrect (1,1,5,5)
$right_bank = selection:fillrect (7,1,11,5)
IF [33%] {
    OBJECT:('!',"levitation"),rndcoord($left_bank),blessed
} ELSE {
    IF [50%] {
        OBJECT:('=',"levitation"),rndcoord($left_bank),blessed
    } ELSE {
        OBJECT:('[',"levitation boots"),rndcoord($left_bank),blessed
    }
}
BRANCH:(1,1,5,5),(0,0,0,0)
STAIR:rndcoord($right_bank),down
"""
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLC(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="lava_crossing.des", **kwargs)


register(
    id="MiniHack-LavaCross-Levitate-Potion-Pickup-v0",
    entry_point="minihack.envs.skills_lava:MiniHackLCLevitatePotionPickup",
)
register(
    id="MiniHack-LavaCross-Levitate-Potion-Inv-v0",
    entry_point="minihack.envs.skills_lava:MiniHackLCLevitatePotionInv",
)
register(
    id="MiniHack-LavaCross-Levitate-Ring-Pickup-v0",
    entry_point="minihack.envs.skills_lava:MiniHackLCLevitateRingPickup",
)
register(
    id="MiniHack-LavaCross-Levitate-Ring-Inv-v0",
    entry_point="minihack.envs.skills_lava:MiniHackLCLevitateRingInv",
)
register(
    id="MiniHack-LavaCross-Levitate-v0",
    entry_point="minihack.envs.skills_lava:MiniHackLCLevitate",
)
register(
    id="MiniHack-LavaCross-v0",
    entry_point="minihack.envs.skills_lava:MiniHackLC",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn
from torch.nn.modules import transformer


class LearnedPositionalEncoder(nn.Module):
    def __init__(self, k, height, width, device):
        super().__init__()

        self.height = height
        self.width = width

        self.enc = torch.randn(height, width, k)

        self.enc = self.enc.div(
            torch.norm(self.enc, p=2, dim=2)[:, :, None].expand_as(self.enc)
        )

        self.mlp = nn.Sequential(
            nn.Linear(2 * k, k), nn.ReLU(), nn.Linear(k, k), nn.ReLU()
        )

        self.enc = nn.Parameter(self.enc, requires_grad=True)[None, :, :, :]

        if device is not None:
            self.enc = self.enc.to(device)

    def forward(self, x):
        x = torch.cat([x, self.enc.expand_as(x)], dim=3)
        x = self.mlp(x)
        return x


class TransformerEncoder(nn.Module):
    def __init__(self, d_model, N, heads, height, width, device):
        super().__init__()
        self.N = N
        self.pe = LearnedPositionalEncoder(d_model, height, width, device)
        self.layers = transformer._get_clones(
            transformer.TransformerEncoderLayer(
                d_model, heads, dim_feedforward=d_model
            ),
            N,
        )

    def forward(self, src, mask=None):
        x = src
        x = self.pe(x)

        bs, h, w, k = x.shape

        x = x.view(bs, h * w, k).transpose(1, 0)

        for i in range(self.N):
            x = self.layers[i](x, mask)

        # FIXME: probably slow due to contiguous; we can adapt the rest of the base
        # model to not assume the batch as first dimension
        return x.transpose(1, 0).view(bs, h, w, k).contiguous()



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation
from minihack.envs import register
from nle import nethack

MOVE_ACTIONS = tuple(nethack.CompassDirection)
NAVIGATE_ACTIONS = MOVE_ACTIONS + (
    nethack.Command.OPEN,
    nethack.Command.KICK,
    nethack.Command.SEARCH,
)


class MiniHackCorridor(MiniHackNavigation):
    """Environment for "corridor" task.

    The agent has to navigate itself through randomely generated corridors that
    connect several rooms and find the goal.
    """

    def __init__(self, *args, des_file, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 1000)
        kwargs["actions"] = NAVIGATE_ACTIONS
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackCorridor2(MiniHackCorridor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="corridor2.des", **kwargs)


class MiniHackCorridor3(MiniHackCorridor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="corridor3.des", **kwargs)


class MiniHackCorridor5(MiniHackCorridor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="corridor5.des", **kwargs)


register(
    id="MiniHack-Corridor-R2-v0",
    entry_point="minihack.envs.corridor:MiniHackCorridor2",
)
register(
    id="MiniHack-Corridor-R3-v0",
    entry_point="minihack.envs.corridor:MiniHackCorridor3",
)
register(
    id="MiniHack-Corridor-R5-v0",
    entry_point="minihack.envs.corridor:MiniHackCorridor5",
)



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation, LevelGenerator
from minihack.envs import register


class MiniHackFightCorridor(MiniHackNavigation):
    def __init__(self, *args, lit=True, **kwargs):
        kwargs["character"] = "kni-hum-law-fem"  # tested on human knight
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 350)

        map = """
-----       ----------------------
|...|       |....................|
|....#######.....................|
|...|       |....................|
-----       ----------------------
"""
        lvl_gen = LevelGenerator(map=map, lit=lit)
        lvl_gen.set_start_rect((1, 1), (3, 3))
        lvl_gen.add_monster(name="giant rat", place=(30, 1))
        lvl_gen.add_monster(name="giant rat", place=(30, 2))
        lvl_gen.add_monster(name="giant rat", place=(30, 3))
        lvl_gen.add_monster(name="giant rat", place=(31, 1))
        lvl_gen.add_monster(name="giant rat", place=(31, 2))
        lvl_gen.add_monster(name="giant rat", place=(31, 3))
        lvl_gen.add_goal_pos((32, 2))

        super().__init__(*args, des_file=lvl_gen.get_des(), **kwargs)


class MiniHackFightCorridorDark(MiniHackFightCorridor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, lit=False, **kwargs)


register(
    id="MiniHack-CorridorBattle-v0",
    entry_point="minihack.envs.fightcorridor:MiniHackFightCorridor",
)

register(
    id="MiniHack-CorridorBattle-Dark-v0",
    entry_point="minihack.envs.fightcorridor:MiniHackFightCorridorDark",
)



# Copyright (c) Facebook, Inc. and its affiliates.

import minihack  # noqa
from gym import envs

skip_envs_list = ["MiniHack-Navigation-Custom-v0", "MiniHack-Skill-Custom-v0"]


def main():
    all_envs = envs.registry.all()
    env_ids = [
        env_spec.id
        for env_spec in all_envs
        if env_spec.id.startswith("MiniHack")
        and env_spec.id not in skip_envs_list
    ]
    print("\n".join(env_ids))


if __name__ == "__main__":
    main()
